#!/usr/bin/env python
# -*- mode: python; coding: koi8-r -*-

import sys, string, gzip, os
import formatter, htmllib
try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO
import base64

from utils import string_encode, error_dialog
from config import default_description_length, default_encoding, preview_length
from bookmarks import Book
from detectcharset import detectcharset
import filters


##----------------------------------------------------------------------
##----------------------------------------------------------------------

class basic_parser:
    def new_book(self, filename):
        """Create new book object"""
        pass
    def load_file(self, buffer, book):
        """Load file in buffer"""
        pass
    def preview(self, buffer, book, offset):
        """Create preview: load part of file in buffer"""
        pass

##----------------------------------------------------------------------
##----------------------------------------------------------------------

class plain_text_parser:

##----------------------------------------------------------------------

    def new_book(self, filename, category=None,
                 external_filter=None, encoding=None):
        """Create new book object"""

        book = Book()
        book.filename = filename

        if external_filter:
            book.externalfilter = external_filter
            if not encoding:
                fileobj = StringIO(self._read_from_pipe(filename,
                                                        external_filter))
                encoding = detectcharset(fileobj)
        else:
            fileobj = open(filename)
            encoding = detectcharset(fileobj)

        if encoding:
            book.encoding = encoding
        else:
            book.encoding = default_encoding

        book.file_type = 'plain_text'
        book.description = self.book_description(book)
        if category:
            book.categories = [category]

        return book

##----------------------------------------------------------------------

    def load_file(self, buffer, book, bookmarks):
        """Load file in buffer"""

        data = self._read_file(filename=book.filename,
                               int_filters=book.internalfilters,
                               ext_filter=book.externalfilter,
                               charset=book.encoding, bookmarks=bookmarks)

        buffer.set_text(data)

##----------------------------------------------------------------------

    def preview(self, buffer, book, offset):
        """Create preview: load part of file in buffer"""

        data = self._read_file(filename=book.filename,
                               offset=offset, length=preview_length,
                               charset=book.encoding,
                               int_filters=book.internalfilters,
                               ext_filter=book.externalfilter)

        buffer.set_text(data)

##----------------------------------------------------------------------

    def binary_content(self, filename):
        return []

##----------------------------------------------------------------------

    def _read_from_pipe(self, filename, command):

        if command.find('%s') > 0:
            command = command % filename
        else:
            command = command + ' ' + filename
        stdin, stdout, stderr = os.popen3(command)
        data = stdout.read()
        err = stderr.read()
        if err:
            raise IOEroor, err
            #error_dialog(None, '', err)

        return data

##----------------------------------------------------------------------

    def _read_file(self, filename, charset, offset=None, length=None,
                   int_filters=None, ext_filter=None, bookmarks=None):

        if ext_filter:
            data = self._read_from_pipe(filename, ext_filter)

            #pipe = os.popen(command)
            #data = pipe.read()
            #pipe.close()

        else:

            fd = open(filename)
            data = fd.read()
            fd.close()

        data = self._get_slice(data, charset, offset, length,
                               int_filters, bookmarks)

        return data

##----------------------------------------------------------------------
    # FIXME: rename
    def _get_slice(self, data, charset, offset=None, length=None,
                   int_filters=None, bookmarks=None):

        # delete invalid (\0) chars
        tbl = string.maketrans('', '')
        data = string.translate(data, tbl, '\0')

        data = unicode(data, charset, 'replace')
        if offset or length:
            if offset:
                data = data[offset:]
            if length:
                data = data[:length]
        data = data.encode('utf-8')

        if int_filters:
            for fltr in int_filters:
                func = getattr(filters, fltr)
                data = func(data, bookmarks=bookmarks)

        return data

##----------------------------------------------------------------------

    def book_description(self, book):
        """Create description of book"""

        data = self._read_file(filename=book.filename, charset=book.encoding,
                               length=default_description_length,
                               ext_filter=book.externalfilter)

        data += ' ...'

        return data

##----------------------------------------------------------------------
##----------------------------------------------------------------------

class gzipped_plain_text_parser(plain_text_parser):

    def _read_file(self, filename, charset, offset=None, length=None,
                   int_filters=None, ext_filter=None, bookmarks=None):

        fd = gzip.open(filename)
        data = fd.read()
        fd.close()

        data = self._get_slice(data, charset, offset, length,
                               int_filters, bookmarks)

        return data

##----------------------------------------------------------------------

    def new_book(self, filename, category=None,
                 external_filter=None, encoding=None):
        """Create new book object"""

        book = Book()
        book.filename = filename

        fileobj = gzip.open(filename)
        encoding = detectcharset(fileobj)
        if encoding:
            book.encoding = encoding
        else:
            book.encoding = default_encoding

        book.file_type = 'gzipped_plain_text'
        book.description = self.book_description(book)
        if category:
            book.categories=[category]

        return book

##----------------------------------------------------------------------
##----------------------------------------------------------------------

class zip_parser(plain_text_parser):

    def _read_file(self, filename, charset, offset=None, length=None,
                   int_filters=None, ext_filter=None, bookmarks=None):

        fd = gzip.open(filename)
        data = fd.read()
        fd.close()

        data = self._get_slice(data, charset, offset, length,
                               int_filters, bookmarks)

        return data

##----------------------------------------------------------------------

    def new_book(self, filename, category=None,
                 external_filter=None, encoding=None):
        """Create new book object"""

        book = Book()
        book.filename = filename

        fileobj = gzip.open(filename)
        encoding = detectcharset(fileobj)
        if encoding:
            book.encoding = encoding
        else:
            book.encoding = default_encoding

        book.file_type = 'zipped_plain_text'
        book.description = self.book_description(book)

        return book

##----------------------------------------------------------------------
##----------------------------------------------------------------------

class MyDumbWriter(formatter.NullWriter):

    def __init__(self):
        self.data = []
        self.bookmarks = []
        self.current_bookmark = 0
        self.index = 0
        formatter.NullWriter.__init__(self)

    def send_paragraph(self, blankline):
        s = '\n'*blankline
        self.data.append(s)
        self.index += len(s)

    def send_line_break(self):
        self.data.append('\n')
        self.index += 1

    def send_hor_rule(self, *args, **kw):
        self.data.append('\n')
        #self.data.append('\n')
        self.index += 1

    def send_literal_data(self, data):
        self.data.append(data)
        self.index += len(data)

    def send_flowing_data(self, data):
        if not data: return
        self.data.append(data)
        self.index += len(data)

    def new_font(self, font):
        # add headers (<h1> <h2> ... ) to bookmarks
        if font and font[0]: # begin of header
            self.current_bookmark_offset = self.index
            self.current_bookmark = 1
            return
        if self.current_bookmark: # end of header
            self.current_bookmark = 0
            self.bookmarks.append((self.current_bookmark_offset, self.index))


##----------------------------------------------------------------------

class MyHTMLParser(htmllib.HTMLParser):
    def __init__(self, formatter):
        htmllib.HTMLParser.__init__(self, formatter)
        self.anchor_list_offset = []

    def anchor_bgn(self, href, name, type):
        self.anchor = href
        if self.anchor:
            self.anchorlist.append(href)

    def anchor_end(self):
        if self.anchor:
            self.handle_data("[%d]" % len(self.anchorlist))
            self.anchor = None

##----------------------------------------------------------------------
##----------------------------------------------------------------------

class html_parser(plain_text_parser):

    def new_book(self, filename, category=None,
                 external_filter=None, encoding=None):
        """Create new book object"""

        book = Book()
        book.filename = filename

        if external_filter:
            book.externalfilter = external_filter
            fileobj = StringIO(self._read_from_pipe(filename, external_filter))
            if not encoding:
                encoding = detectcharset(fileobj)
        else:
            fileobj = open(filename)
            encoding = detectcharset(fileobj)

        if encoding:
            book.encoding = encoding
        else:
            encoding = book.encoding = default_encoding

        book.file_type = 'html'
        book.description = self.book_description(book)
        if category:
            book.categories = [category]

        return self._html_new_book(fileobj, book, encoding)

##----------------------------------------------------------------------

    def _html_new_book(self, fileobj, book, encoding):

        fileobj.seek(0)
        ex = fileobj.read()
        w = MyDumbWriter()
        h = htmllib.HTMLParser(formatter.AbstractFormatter(w))
        h.feed(ex)
        data = ''.join(w.data)
        if h.title:
            book.name = unicode(h.title, encoding, 'replace').encode('utf-8')

        text_len = len(data)
        data = unicode(data, encoding, 'replace')

        for bm in w.bookmarks:
            desc = data[bm[0]:bm[1]]
            desc = desc.encode('utf-8')
            book.add_bookmark(float(bm[0])/text_len, bm[0], bm[1]-bm[0], desc)

        return book

##----------------------------------------------------------------------

    def _read_file(self, filename, charset, offset=None, length=None,
                   int_filters=None, ext_filter=None, bookmarks=None):

        if ext_filter:
            data = self._read_from_pipe(filename, ext_filter)
        else:
            data = open(filename).read()

        return self._parse_html(data, charset, offset, length,
                                int_filters, ext_filter, bookmarks)

#----------------------------------------------------------------------

    def _parse_html(self, ex, charset, offset, length,
                    int_filters, ext_filter, bookmarks):

        w = MyDumbWriter()
        h = htmllib.HTMLParser(formatter.AbstractFormatter(w))
        h.feed(ex)
        data = string.join(w.data, '')
        if len(h.anchorlist) > 0:
            # add links
            data += '\n\nLinks:\n'
            n = 1
            for i in h.anchorlist:
                data += '  [%d] %s\n' % (n, i)
                n += 1

        data = self._get_slice(data, charset, offset, length,
                               int_filters, bookmarks)

        return data

##----------------------------------------------------------------------

class gzipped_html_parser(html_parser):

    def new_book(self, filename, category=None,
                 external_filter=None, encoding=None):
        """Create new book object"""

        book = Book()
        book.filename = filename

        fileobj = gzip.open(filename)
        encoding = detectcharset(fileobj)

        if encoding:
            book.encoding = encoding
        else:
            encoding = book.encoding = default_encoding

        book.file_type = 'gzipped_html'
        book.description = self.book_description(book)
        if category:
            book.categories = [category]

        return self._html_new_book(fileobj, book, encoding)

##----------------------------------------------------------------------

    def _read_file(self, filename, charset, offset=None, length=None,
                   int_filters=None, ext_filter=None, bookmarks=None):

        fd = gzip.open(filename)
        data = fd.read()
        fd.close()

        return self._parse_html(data, charset, offset, length,
                                int_filters, ext_filter, bookmarks)

##----------------------------------------------------------------------
##----------------------------------------------------------------------
##----------------------------------------------------------------------

class LibRuWriter(formatter.NullWriter):

    def __init__(self):
        self.data = []
        self.bookmarks = []
        formatter.NullWriter.__init__(self)

    def send_paragraph(self, blankline):
        self.data.append('\n'*blankline)

    def send_line_break(self):
        pass
        #self.data.append('\n')

    def send_hor_rule(self, *args, **kw):
        self.data.append('\n')
        #self.data.append('\n')

    def send_literal_data(self, data):
        self.data.append(data)

    def send_flowing_data(self, data):
        if not data: return
        if data[0] == ' ':
            self.data.append('\t')
        for i in data.strip().split():
            self.data.append(i+' ')

    def new_font(self, font):
        if font and font[0]:
            #self.bookmarks.append(len(' '.join(self.data)))
            self.bookmarks.append(reduce(lambda x, y: x+len(y), self.data, 0))

#----------------------------------------------------------------------

class lib_ru_parser(html_parser):

    def _parse_html(self, ex, charset, offset, length,
                    int_filters, ext_filter, bookmarks):

        w = LibRuWriter()
        h = htmllib.HTMLParser(formatter.AbstractFormatter(w))
        h.feed(ex)
        data = ''.join(w.data)
        if len(h.anchorlist) > 0:
            data += '\n\nLinks:\n'
            n = 1
            for i in h.anchorlist:
                data += '  [%d] %s\n' % (n, i)
                n += 1

        data = self._get_slice(data, charset, offset, length,
                               int_filters, bookmarks)

        return data


##----------------------------------------------------------------------
##----------------------------------------------------------------------

class fb2_parser(plain_text_parser):

    def new_book(self, filename, category=None,
                 external_filter=None, encoding=None):
        """Create new book object"""
        from fictionbook2 import FictionBook2

        book = Book()
        book.filename = filename
        book.encoding = 'utf-8'
        book.file_type = 'fb2'

        fb = FictionBook2(filename)

        book.author = fb.author
        book.name = fb.name
        book.description = fb.description
        text = fb.text
        u_text = text.decode('utf-8')
        text_len = len(u_text)
        for b in fb.bookmarks:
            #pass
            offset = b[0]
            length = b[1]
            book.add_bookmark(float(b[0])/text_len, offset, length,
                              u_text[b[0]:b[0]+b[1]].encode('utf-8'))

        book.categories = book.genres = fb.genres

        return book

##----------------------------------------------------------------------

    def _read_file(self, filename, charset, offset=None, length=None,
                   int_filters=None, ext_filter=None, bookmarks=None):
        from fictionbook2 import FictionBook2

        fb = FictionBook2(filename)
        data = self._get_slice(fb.text, 'utf-8', offset, length,
                               int_filters, bookmarks)

        return data

##----------------------------------------------------------------------

    def binary_content(self, arg):
        from fictionbook2 import FictionBook2
        import types

        content = []

        if type(arg) is types.StringType:
            fb = FictionBook2(arg)
        else:
            fb = arg

        for bin_id, bin_ct, bin_data, bin_offset in fb.binaries:
            try:
                data=base64.decodestring(bin_data)
            except Exception, err:
                print >> sys.stderr, \
                      'Error in base64.decodestring: id: %s: %s' \
                      % (bin_id, err)
                continue
            content.append((bin_id, bin_ct, data, bin_offset))
        return content

##----------------------------------------------------------------------

    def load_file(self, buffer, book, bookmarks):
        """Load file in buffer"""
        from fictionbook2 import FictionBook2
        import tempfile
        import gtk

        fb = FictionBook2(book.filename)
        self._fb = fb
        text = fb.text

##         data = self._read_file(filename=book.filename,
##                                int_filters=book.internalfilters,
##                                ext_filter=book.externalfilter,
##                                charset=book.encoding, bookmarks=bookmarks)

        buffer.set_text(text)

        for mark in fb.bookmarks:
            start, length = mark
            end = start + length
            iter_start = buffer.get_iter_at_offset(start)
            iter_end = buffer.get_iter_at_offset(end)
            buffer.apply_tag(buffer.title_tag, iter_start, iter_end)

        for style in fb.styles:
            start, length, type = style
            end = start + length
            if type in ('strong', 'emphasis'):
                iter_start = buffer.get_iter_at_offset(start)
                iter_end = buffer.get_iter_at_offset(end)
                if type == 'strong':
                    buffer.apply_tag(buffer.strong_tag, iter_start, iter_end)
                else:
                    buffer.apply_tag(buffer.emphasis_tag, iter_start, iter_end)

        binary_content = self.binary_content(fb)
        for bin_id, bin_ct, bin_data, bin_offset in binary_content:

            if bin_offset < 0: continue

            tmp_file = tempfile.mktemp()
            try:
                open(tmp_file, 'w').write(bin_data)
                pb = gtk.gdk.pixbuf_new_from_file(tmp_file)

                start = buffer.get_iter_at_offset(bin_offset)
                end = buffer.get_iter_at_offset(bin_offset+1)
                buffer.delete(start, end)
                iter = buffer.get_iter_at_offset(bin_offset-1)
                buffer.insert_pixbuf(iter, pb)

                # place pixmap to center
                start = buffer.get_iter_at_offset(bin_offset-1)
                end = buffer.get_iter_at_offset(bin_offset)
                buffer.apply_tag(buffer.pixmap_tag, start, end)

                #pixmap, mask = pb.render_pixmap_and_mask()
                #self.binary_image.set_from_pixmap(pixmap, mask)

            finally:
                os.remove(tmp_file)

        for lnk in fb.links:
            start, end, link_start, link_end, type = lnk
            iter_start = buffer.get_iter_at_offset(start)
            iter_end = buffer.get_iter_at_offset(end)
            if type == 'link_type':
                buffer.apply_tag(buffer.link_tag, iter_start, iter_end)
            else: # note_type
                buffer.apply_tag(buffer.note_tag, iter_start, iter_end)




##----------------------------------------------------------------------

    def goto_link(self, offset):

        for lnk in self._fb.links:
            start, end, link_start, link_end, type = lnk
            if start<=offset<=end:
                return link_start
                return len(unicode(self._fb.text[:link_start], 'utf-8'))


